/**
 * ShimmerProcessor.java
 * Copyright (C) 2012 Pattern Recognition Lab, University Erlangen-Nuremberg.
 *  
 * Licensed under the Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License (the "License");
 * you may not use this file except in compliance with the License.
 * To view a copy of this License, visit http://creativecommons.org/licenses/by-nc-sa/3.0/.
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on 
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the 
 * specific language governing permissions and limitations under the License.
 * 
 * 
 * This file is part of the "Hearty" Android Application. It was released as supplementary material related to the publication [1]:
 * [1] S. Gradl, P. Kugler, C. Lohmüller, and B. Eskofier, “Real-time ECG monitoring and arrhythmia detection using Android-based mobile devices,” in 34th Annual International Conference of the IEEE EMBS, 2012, pp. 2452–2455.
 * 
 * It is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * If you reuse this code you have to keep or cite this comment.
 *
 */
package de.lme.hearty;


import java.util.ArrayList;

import android.os.Handler;
import android.os.Message;



/**
 * @author Patrick Kugler
 */
abstract class ShimmerProcessor implements Handler.Callback, ShimmerDataSource.OnShimmerDataListener
{

	private static final int				MESSAGE_READ	= 0;
	private Handler							messageHandler;
	protected ShimmerDataSource[]			shimmerDevices;
	private OnShimmerProcessorEventListener	mmOnShimmerProcessorEventListener;


	public static interface OnShimmerProcessorEventListener
	{
		abstract void onShimmerProcessorEvent ();
	}


	public ShimmerProcessor (ArrayList< ShimmerDataSource > shimmerDataSources)
	{
		this.messageHandler = new Handler( this );
		this.shimmerDevices = (ShimmerDataSource[]) shimmerDataSources.toArray( new ShimmerDataSource[ 0 ] );
		this.mmOnShimmerProcessorEventListener = null;
	}


	public void setOnShimmerProcessorEventListener (OnShimmerProcessorEventListener listener)
	{
		mmOnShimmerProcessorEventListener = listener;
	}


	public void connectShimmer ()
	{
		for (int i = 0; i < shimmerDevices.length; i++)
		{
			shimmerDevices[ i ].addOnShimmerDataListener( this );
			shimmerDevices[ i ].connect();
		}

	}


	public void disconnectShimmer ()
	{
		for (int i = 0; i < shimmerDevices.length; i++)
		{
			shimmerDevices[ i ].disconnect();
			shimmerDevices[ i ].removeOnShimmerDataListener( this );
		}
	}


	public void startProcessing ()
	{
		for (int i = 0; i < shimmerDevices.length; i++)
		{
			shimmerDevices[ i ].start();
		}
	}


	public void stopProcessing ()
	{
		for (int i = 0; i < shimmerDevices.length; i++)
		{
			shimmerDevices[ i ].stop();
		}
	}


	public void scheduleProcessing ()
	{
		new Thread( new Runnable() {
			public void run ()
			{
				try
				{
					onProcess();
				}
				catch (Exception e)
				{
					return;
				}
			}
		} ).start();
	}


	public boolean handleMessage (Message msg)
	{
		switch (msg.what)
		{
			case MESSAGE_READ:
				mmOnShimmerProcessorEventListener.onShimmerProcessorEvent();
		}

		return true;
	}


	protected void sendNotification ()
	{
		messageHandler.obtainMessage( MESSAGE_READ, -1, -1, null ).sendToTarget();
	}


	protected abstract void onProcess ();
}
